module CliWrap (
		Cli (..), toCliWithLog, cliDoVerbose, cliDoLog, 
		getCliFiles, getCliLogfile, getCliStartStop
	       ) where

import GetOpt (getOpt, usageInfo, ArgOrder(..), OptDescr(..), ArgDescr(..))
import LogResult (LogResult, mapLR)
import List (find)
import Utils (numberLines, isNothing, fromJust, (?), (==>))

data CliOpt = CliVerbose | CliVersion | CliHelp | CliNoLogOutput | CliLog CliLogFile | CliOutput CliOutFile |
	       CliStartActions String | CliStopActions String  deriving (Show, Eq)
data CliOutFile = ODEFAULT | Outfile String deriving (Show, Eq)
data CliLogFile = LDEFAULT | Logfile String deriving (Show, Eq)

type CliVersion = String
type CliArg = String
data Cli = CliValid [CliOpt] CliArg | CliInfo [CliOpt] [CliArg] String | CliError [CliOpt] [CliArg] String

cliOptions :: [OptDescr CliOpt]
cliOptions =
   [Option "v"     ["verbose"]     (NoArg CliVerbose)        "logfile output to STDOUT",
    Option "V"     ["version"]     (NoArg CliVersion)        "show version info",
    Option "h"     ["help"]        (NoArg CliHelp)           "show options and documentation",
    Option "n"     ["nolog"]       (NoArg CliNoLogOutput)    "suppress log file output",
    Option "l"     ["logfile"]     (OptArg cliLog "LOGFILE") "use specified LOGFILE for log output",
    Option "o"     ["output"]      (OptArg cliOut "FILE")    "use specified FILE for output",
    Option "s"     ["source"]      (ReqArg CliStartActions "STARTACTIONS")   "start actions for stochastic probe",
    Option "t"     ["target"]      (ReqArg CliStopActions "STOPACTIONS")     "stop actions for stochastic probe"]

extraInfo :: String
extraInfo = unlines ["Logging: generating log information (either to a file or to STDOUT with -v)",
		     "can more than double the time taken to run ipc. Using the -n option will",
		     "speed up compilation of large models.\n",
		     "STARTACTIONS and STOPACTIONS: these are comma separated lists of action labels",
		     "as they appear in the .pepa file. Used to specify the observed behaviour",
		     "that should start and stop the passage-time measurement."]

cliOut, cliLog :: Maybe String -> CliOpt
cliOut Nothing  = CliOutput $ ODEFAULT
cliOut (Just o) = CliOutput $ Outfile o
cliLog Nothing  = CliLog $ LDEFAULT 
cliLog (Just o) = CliLog $ Logfile o

toCli :: CliVersion -> String -> [CliArg] -> Cli
toCli vers prog xss = let 
			     order = Permute
			     opts = cliOptions
			     header = "Usage: " ++ prog ++ " [OPTION...] [PEPAFILE]"
			     info = "Try `ipc --help' for more information."
			     usage  = (usageInfo header opts) ++ "\n"++ extraInfo
			   in case getOpt order opts xss of
			(os,as,[])   -> processCliOpts (prog, vers) (unlines [header,info],usage) os as
                        (os,as,errs) -> CliError os as (concat ([prog, ": "] ++ errs) ++ "\n" ++ usage)

toCliWithLog :: CliVersion -> String -> LogResult [CliArg] -> LogResult Cli
toCliWithLog vs pg = mapLR "CliWrap" (toCli vs pg) analyseCli

analyseCli :: Cli -> String
analyseCli (CliValid os a) = concat ["Valid processing command line directive parsed.\n",
				    show $ length os, " option(s) and 1 argument found:\n",
				    numberLines $ map show os, "Argument: ", a]
analyseCli (CliInfo os as str)  = concat [show $ length os, " option(s) found:\n", numberLines $ map show os, 
					 show $ length as, " arguments(s) found:\n", numberLines as, 
					 "\nExtra information:\n", str]
analyseCli (CliError os as str) = concat [show $ length os, " option(s) found:\n", numberLines $ map show os, 
					 show $ length as, " arguments(s) found:\n", numberLines as, 
					 "\nError message:\n", str]

processCliOpts :: (String,String) -> (String,String) -> [CliOpt] -> [CliArg] -> Cli
processCliOpts (pr,vs) (shortUsg,usg) os as 
    | isInCliO (CliVersion==) os = CliInfo os as $ concat [pr, " v", vs]
    | isInCliO (CliHelp==) os    = CliInfo os as usg
    | null as                    = CliError os as $ concat [pr,": Input file missing\n\n", shortUsg]
    | length as > 1              = CliError os as $ concat [pr,": Too many input files:\n", foldr1 (\x -> \y -> x++", "++y) as, "\n\n", shortUsg]
    | testNoStartNoStop os       = CliError os as $ concat [pr,": Missing start/stop action list for passage definition\n\n", shortUsg]
    | otherwise                  = CliValid os (head as) where
				 testNoStartNoStop :: [CliOpt] -> Bool
				 testNoStartNoStop os = not ((isInCliO cliIsStart os) && (isInCliO cliIsStop os))
				 

isInCliO :: (CliOpt -> Bool) -> [CliOpt] -> Bool
isInCliO f = or . (map f)
isInCli :: (CliOpt -> Bool) -> Cli -> Bool
isInCli f (CliValid os _) = isInCliO f os
isInCli f (CliInfo os _ _) = isInCliO f os
isInCli f (CliError os _ _) = isInCliO f os

cliDoLog, cliDoVerbose :: Cli -> Bool
cliDoLog = not . (isInCli (CliNoLogOutput==))
cliDoVerbose = isInCli (CliVerbose==)

cliIsOutput, cliIsLogfile, cliIsStart, cliIsStop :: CliOpt -> Bool
cliIsOutput (CliOutput _) = True
cliIsOutput _ = False
cliIsLogfile (CliLog _) = True
cliIsLogfile _ = False
cliIsStart (CliStartActions _) = True
cliIsStart _ = False
cliIsStop (CliStopActions _) = True
cliIsStop _ = False


getCliLogfile :: Cli -> String
getCliLogfile (CliInfo _ _ _)  = error "getCliLogfile: no log file [1]"
getCliLogfile (CliError _ _ _) = error "getCliLogfile: no log file [2]"
getCliLogfile (CliValid os a) 
    | isNothing fd  = def
    | isNothing mlg = def
    | otherwise     = fromJust mlg where
		  mlg = getLFile $ fromJust fd
		  def = ((reverse end) == "pepa") ? ((reverse root)++"log", a++".log")
		  (end, root) = span ('.'/=) (reverse a)
		  fd = find cliIsLogfile os 
		  getLFile :: CliOpt -> Maybe String
		  getLFile (CliLog LDEFAULT) = Nothing
		  getLFile (CliLog (Logfile s)) = Just s
		  getLFile _ = error "getLFile: other CliOpt found"


getCliStartStop :: Cli -> ([String], [String])
getCliStartStop (CliInfo _ _ _)  = error "getCliStartStop: no start/stop actions needed [1]"
getCliStartStop (CliError _ _ _) = error "getCliStartStop: no start/stop actions needed [2]"
getCliStartStop (CliValid os _)  
    | isNothing s1 = error "getCliStartStop: start actions should exist"
    | isNothing s2 = error "getCliStartStop: stop actions should exist"
    | ((null fs1) || (null fs2)) = error "getCliStartStop: null start/stop action"
    | otherwise    = (removeCommas fs1, removeCommas fs2)where
		  [CliStartActions fs1, CliStopActions fs2] = map fromJust [s1, s2]
		  s1 = find cliIsStart os
		  s2 = find cliIsStop os
		  removeCommas :: String -> [String] 
		  removeCommas xs = lines $ map (\x -> (x==',') ? ('\n',x)) xs

getCliFiles :: Cli -> (String, String)
getCliFiles (CliInfo _ _ _)  = error "getCliFiles: no in/out file needed [1]"
getCliFiles (CliError _ _ _) = error "getCliFiles: no in/out file needed [2]"
getCliFiles (CliValid os a)
    | isNothing fd = def
    | isNothing oFile = def
    | otherwise = (fst def, fromJust oFile) where
		  def = ((reverse end) == "pepa") ? ((a,(reverse root)++"mod"), (a++".pepa",a++".mod"))
		  (end, root) = span ('.'/=) (reverse a)
		  oFile = getOFile $ fromJust fd
		  fd = find cliIsOutput os 
		  getOFile :: CliOpt -> Maybe String
		  getOFile (CliOutput ODEFAULT) = Nothing
		  getOFile (CliOutput (Outfile s)) = Just s
		  getOFile _ = error "getOFile: other CliOpt found"

{-
newFilename :: String -> String
newFilename fs = ((reverse end) == "pepa") ? ((reverse root)++"mod", fs++".mod") where
		 (end, root) = span ('.'/=) (reverse fs)
-}
